import uvicorn as uvicorn
from examples.meeting_room.domain_models import User

from examples.meeting_room.view_models import UserView
from onto import view_model, attrs
from onto.view import Mediator


class UserGraphQLMediator(Mediator):

    from onto.source import domain_model
    from onto.sink.graphql import subscription

    # src = domain_model(domain_model_cls=User)
    from onto.source.kafka import KafkaSource
    # src = KafkaSource(topic_name='users')
    from onto.source.mock import MockDomainModelSource
    src = MockDomainModelSource(User)
    subscribe_user_view = subscription(view_model_cls=UserView)

    @subscribe_user_view.triggers.add_topic
    def add_topic(self, user_id: str):
        return user_id

    @subscribe_user_view.triggers.on_event
    def on_event(self, event: dict):
        return event

    @src.triggers.on_create
    async def user_added(self, obj):
        q = self.subscribe_user_view.qs[obj.doc_id]
        d = obj.to_dict()
        await q.put(d)
        # import asyncio
        # # tell asyncio to enqueue the result
        # fut = asyncio.run_coroutine_threadsafe(
        #     , loop=self.subscribe_user_view.loop
        # )
        # # wait for the result to be enqueued
        # _ = fut.result()

    # @src.triggers.on_topic
    # def user_updated(self, message):
    #     # v = UserView.get(user_id=obj.doc_id)
    #     # d = v.to_dict()
    #     print(message)
    #     # raise ValueError(str(message))
    #     # User.from_dict(d=d)
    #     # q = self.subscribe_user_view.qs[obj.doc_id]
    #     #
    #     # import asyncio
    #     # # tell asyncio to enqueue the result
    #     # fut = asyncio.run_coroutine_threadsafe(
    #     #     q.put(d), loop=self.subscribe_user_view.loop
    #     # )
    #     # # wait for the result to be enqueued
    #     # _ = fut.result()

    @classmethod
    def start(cls):
        # cls.src.start()

        import asyncio
        loop = asyncio.get_event_loop()
        cls.src.start(loop=loop)

        s = cls.subscribe_user_view.start()
        from onto.sink.graphql import graph_schema
        import graphql
        liveness = graph_schema(
            op_type='Query',
            name='liveness',
            graphql_object_type=graphql.GraphQLObjectType(
                name='Liveness',
                fields={
                    'alive': graphql.GraphQLField(
                        graphql.GraphQLBoolean,
                        resolve=lambda *args, **kwargs: True),
                }
            ),
            args=dict()
        )

        return [s, liveness]


class LoginForm(view_model.ViewModel):
    phone_number = attrs.bproperty(type_cls=str)
    verification_code = attrs.bproperty(type_cls=str)


from onto.attrs import attrs
class LoginRes(view_model.ViewModel):
    token = attrs.of_type(str)


class UserAddMediator(Mediator):

    from onto.sink.graphql import mutation

    login_user_view = mutation(view_model_cls=LoginRes)

    @login_user_view.triggers.mutate
    def login(self, user_id: str, first_name: str, last_name: str, org: str):
        user = User.new(
            doc_id=user_id, first_name=first_name,
            last_name=last_name, organization=org,
            hearing_aid_requested=True
        )
        user.save()

        return LoginRes.new(
            token='supposed to be token for user added'
        )

    @classmethod
    def start(cls):
        s = cls.login_user_view.start()
        return [s]


class UserLoginMediator(Mediator):

    from onto.sink.graphql import mutation

    login_user_view = mutation(view_model_cls=LoginRes)

    @login_user_view.triggers.mutate
    def login(self, form: LoginForm):
        print(form)

    @classmethod
    def start(cls):
        s = cls.login_user_view.start()
        return [s]





def default_field_resolver(source, info, **args):
    """Default field resolver.

    If a resolve function is not given, then a default resolve behavior is used which
    takes the property of the source object of the same name as the field and returns
    it as the result, or if it's a function, returns the result of calling that function
    while passing along args and context.

    For dictionaries, the field names are used as keys, for all other objects they are
    used as attribute names.
    """
    # Ensure source is a value for which property access is acceptable.
    field_name = info.field_name
    value = (
        source.get(field_name)
        if isinstance(source, dict)
        else source.graphql_field_resolve(info, **args)
    )
    if callable(value):
        return value(info, **args)
    return value


class GraphqlAttributedMixin:
    from graphql import GraphQLResolveInfo

    def graphql_field_resolve(self, info: GraphQLResolveInfo, *args):
        field_name = info.field_name
        return self.graphql_representation.get(field_name, None)

    import functools

    @property
    @functools.lru_cache(maxsize=None)
    def graphql_representation(self):
        return self.to_dict()



from flask import request

from onto.view.base import ViewMediatorBase

from flask_socketio import Namespace, emit, send

from functools import partial

import json


class ViewMediatorWebsocket(ViewMediatorBase, Namespace):
    """
    Watches and updates Firestore for DocumentAsView view models
    """

    def __init__(self,
                 view_model_cls=None,
                 mutation_cls=None, *args, **kwargs):
        """

        :param view_model_cls: the view model to be exposed in REST API
        :param app: Flask App
        :param mutation_cls: a subclass of Mutation to handle the changes
                POST, PATCH, UPDATE, PUT, DELETE made to the list of view
                models or a single view model.
        """
        super().__init__(*args, **kwargs)
        self.view_model_cls = view_model_cls
        self.mutation_cls = mutation_cls
        self.rule_view_cls_mapping = dict()
        self.default_tag = self.view_model_cls.__name__
        self.instances = dict()

    @classmethod
    def notify(cls, self, obj):
        """ Specifies what to do with the view model newly generated
                from an update.

        :param obj:
        :return:
        """
        self.emit("updated", obj.to_view_dict())

    def on_connect(self):
        """
        Ref: https://github.com/miguelgrinberg/Flask-SocketIO/blob/master/test_socketio.py
        :return:
        """
        if request.args.get('fail'):
            return False
        send('connected')
        send(json.dumps(request.args.to_dict(flat=False)))
        send(json.dumps({h: request.headers[h] for h in request.headers.keys()
                         if
                         h not in ['Host', 'Content-Type', 'Content-Length']}))

    def on_disconnect(self):
        pass

    def on_subscribe_view_model(self, data):
        emit("subscribed")
        self.instances[0] = self.view_model_cls.new(
            **data,
            once=False,
            f_notify=partial(self.notify, self=self)
        )



from flasgger import Swagger
from flask import Flask

from examples.meeting_room.view_models import MeetingSession, UserView
from examples.meeting_room.view_models.meeting_session import \
    MeetingSessionMutation
from examples.meeting_room.views import meeting_session_ops
from onto.view import rest_api

app = Flask(__name__)

meeting_session_mediator = rest_api.ViewMediator(
    view_model_cls=MeetingSession,
    app=app,
    mutation_cls=MeetingSessionMutation
)
meeting_session_mediator.add_list_get(
    rule="/meeting_sessions",
    list_get_view=meeting_session_ops.ListGet
)

meeting_session_mediator.add_instance_get(
    rule="/meeting_sessions/<string:doc_id>")
meeting_session_mediator.add_instance_patch(
    rule="/meeting_sessions/<string:doc_id>")

user_mediator = rest_api.ViewMediator(
    view_model_cls=UserView,
    app=app,
)
user_mediator.add_instance_get(
    rule="/users/<string:doc_id>"
)

swagger = Swagger(app)

if __name__ == "__main__":
    app.run(debug=True)



import leancloud
from . import Database, Reference, Snapshot
from ..common import _NA
from ..query.query import Query

LEANCLOUD_DOC_ID_DATA_KEY = '_doc_id'

class LeancloudDatabase(Database):

    class Comparators(Database.Comparators):

        eq = 'equal_to'
        gt = 'greater_than'
        ge = 'greater_than_or_equal_to'
        lt = 'less_than'
        le = 'less_than_or_equal_to'

        _in = 'contained_in'

    @classmethod
    def _get_cla(cls, cla_str):
        return leancloud.Object.extend(name=cla_str)

    ref = Reference()

    @classmethod
    def set(cls, ref: Reference, snapshot: Snapshot, transaction=_NA):
        object_id = ref.last
        cla = cls._get_cla(ref.first)
        obj = cla(_doc_id=object_id, **snapshot.to_dict())
        obj.save()

    @classmethod
    def get(cls, ref: Reference, transaction=_NA):
        _doc_id = ref.last
        cla = cls._get_cla(ref.first)
        cla_obj = cla.query.equal_to(LEANCLOUD_DOC_ID_DATA_KEY, _doc_id).first()
        d = LeancloudSnapshot.from_cla_obj(cla_obj).to_dict()
        d['doc_id'] = d.pop(LEANCLOUD_DOC_ID_DATA_KEY)
        snapshot = LeancloudSnapshot(d)
        return snapshot

    @classmethod
    def update(cls, ref: Reference, snapshot: Snapshot, transaction=_NA):
        cls.set(ref, snapshot, transaction)

    @classmethod
    def create(cls, ref: Reference, snapshot: Snapshot, transaction=_NA):
        cls.set(ref, snapshot, transaction)

    @classmethod
    def delete(cls, ref: Reference, transaction=_NA):
        """ Note: this only deletes one instance that has _doc_id == ref.last

        :param ref:
        :param transaction:
        :return:
        """
        _doc_id = ref.last
        cla = cls._get_cla(ref.first)
        cla_obj = cla.query.equal_to(LEANCLOUD_DOC_ID_DATA_KEY, _doc_id).first()
        cla_obj.destroy()

    @classmethod
    def query(cls, q: Query):
        for cla_obj in q._to_leancloud_query().find():
            ref = LeancloudReference.from_cla_obj(cla_obj)
            snapshot = LeancloudSnapshot.from_cla_obj(cla_obj)
            yield (ref, snapshot)


class LeancloudSnapshot(Snapshot):
    """
    TODO: apply to LeancloudDatabase class and other places
    """

    @classmethod
    def from_cla_obj(self, cla_obj):
        snapshot = Snapshot(cla_obj.dump())
        return snapshot


class LeancloudReference(Reference):

    @classmethod
    def from_cla_obj(cls, cla_obj):
        ref = cls()/cla_obj._class_name/cla_obj.get('_doc_id', cla_obj.id)
        return ref



"""
Note that test setup is UNRELIABLE.
Class may persist in registry (of metaclass) across tests.

"""
import json

import pytest as pytest
from functools import lru_cache

import onto.models.base
from onto import view_model, schema, fields


@pytest.fixture
@lru_cache(maxsize=1)
def ModelASchema():
    class ModelASchema(schema.Schema):
        int_a = fields.Integer(load_from="intA", dump_to="intA")
        int_b = fields.Integer(load_from="intB", dump_to="intB")

    return ModelASchema


@pytest.fixture
@lru_cache(maxsize=1)
def ModelA(ModelASchema):
    class ModelA(onto.models.base.Serializable):
        _schema_cls = ModelASchema

    return ModelA


def test_create_model():
    class ModelKSchema(schema.Schema):
        int_a = fields.Integer(load_from="intA", dump_to="intA")
        int_b = fields.Integer(load_from="intB", dump_to="intB")

    class ModelK(onto.models.base.Serializable):
        class Meta:
            schema_cls = ModelKSchema

    obj: ModelK = ModelK.new()

    assert obj.int_a == 0
    assert obj.int_b == 0

    assert obj.to_dict() == {
        "intA": 0,
        "intB": 0,
        "obj_type": "ModelK",
    }


def test_cls_factory(ModelA):
    obj: ModelA = ModelA.new()

    assert obj.int_a == 0
    assert obj.int_b == 0

    obj.int_a = 1
    obj.int_b = 2

    assert obj.int_a == 1
    assert obj.int_b == 2

    assert obj.to_dict() == {
        "intA": 1,
        "intB": 2,
        "obj_type": "ModelA",
    }


def test_from_dict(ModelA):
    obj = ModelA.from_dict({
        "intA": 1,
        "intB": 2,
        "obj_type": "ModelA",
    })

    assert isinstance(obj, ModelA)
    assert obj.int_a == 1
    assert obj.int_b == 2


def test__additional_fields(ModelASchema, ModelA):
    class ModelAASchema(ModelASchema):
        int_aa = fields.Integer(load_from="intAA", dump_to="intAA")

    obj_a = ModelA.new()

    assert hasattr(obj_a, "int_a")
    assert not hasattr(obj_a, "int_aa")

    from onto.models.factory import ClsFactory
    ModelAA = ClsFactory.create(
        name="ModelAA",
        schema=ModelAASchema
    )
    obj_aa = ModelAA.new()
    assert hasattr(obj_aa, "int_a")
    assert hasattr(obj_aa, "int_aa")


def test_default_value(ModelA):
    obj_a = ModelA.new()

    assert obj_a.int_a == 0


def test_property_fields():
    class ModelAPSchema(schema.Schema):
        some_property = fields.Integer(dump_only=True)

    def fget(self):
        return 8

    sp = property(fget=fget)

    class ModelAP(onto.models.base.Serializable):
        _schema_cls = ModelAPSchema

    ModelAP.some_property = sp

    obj_a = ModelAP.new()

    assert obj_a.some_property == 8


def test_multiple_inheritance(ModelASchema):
    class ModelBSchema(schema.Schema):
        int_b = fields.Integer(load_from="intB", dump_to="intB")

    class ModelABSchema(ModelASchema, ModelBSchema):
        pass

    from onto.models.factory import ClsFactory
    ModelAB = ClsFactory.create(
        name="ModelAB",
        schema=ModelABSchema
    )

    ab = ModelAB.new()
    ab.int_a = 1
    ab.int_b = 2

    assert ab.int_a == 1
    assert ab.int_b == 2

    assert ab._export_as_dict().items() >= {
        "intA": 1,
        "intB": 2
    }.items()


def test__export_as_dict(ModelA):
    a = ModelA.new()
    a.int_a = 1
    a.int_b = 2

    assert a._export_as_dict() == {
        "intA": 1,
        "intB": 2,
        "obj_type": "ModelA",
    }


def test__import_properties(ModelA):
    a = ModelA.new()
    a.update_vals(with_dict={
        "int_a": 1,
        "int_b": 2,
    })

    assert a.int_a == 1
    assert a.int_b == 2


def test_separate_class():
    """
    Tests experimental code for decorator pattern implementation

    :return:
    """

    class SModelASchema(schema.Schema):
        int_a = schema.fields.Integer(load_from="intA", dump_to="intA")
        int_b = schema.fields.Integer(load_from="intB", dump_to="intB")

    class SModelASerializable(onto.models.base.Serializable):
        _schema_cls = SModelASchema

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

    class SModelAModel(object):

        def __init__(self):
            self.int_a = 0
            self.int_b = 0

    class SModelA(SModelAModel, SModelASerializable):

        def __init__(self, **kwargs):
            super().__init__()
            super(SModelASerializable, self).__init__(**kwargs)

    a = SModelA.new()
    a.int_a = 1
    a.int_b = 2

    assert a._export_as_dict() == {
        "intA": 1,
        "intB": 2,
        "obj_type": "SModelA",
    }


def test_many():

    class TimeRangeSchema(schema.Schema):
        earliest = fields.Raw()
        latest = fields.Raw()

    class TimeRange(onto.models.base.Serializable):
        _schema_cls = TimeRangeSchema

    t = TimeRange()
    t.earliest = 10
    t.latest = 20

    from marshmallow import fields as m_fields

    class MyPlanSchema(schema.Schema):
        range = m_fields.Nested(TimeRangeSchema)
        ranges = m_fields.List(fields.Embedded(obj_cls='TimeRange'))
        ranged = m_fields.Dict(fields.Embedded(obj_cls='TimeRange'))
        name = fields.Str()

    class MyPlan(onto.models.base.Serializable):
        _schema_cls = MyPlanSchema

    d = MyPlanSchema().load({
        "range": t.to_dict(),
        "ranges": [t.to_dict(), t.to_dict()],
        "ranged": {
          'a': t.to_dict(),
          'b': t.to_dict()
        },
        "name": "my plan"
    })

    assert d != dict()


def test_embedded():
    class TargetSchema(schema.Schema):
        earliest = fields.Raw()
        latest = fields.Raw()

    class Target(onto.models.base.Serializable):
        _schema_cls = TargetSchema

    t = Target()
    t.earliest = 10
    t.latest = 20

    class PlanSchema(schema.Schema):
        target = fields.Embedded(obj_cls="Target")
        name = fields.Str()

    class Plan(onto.models.base.Serializable):
        _schema_cls = PlanSchema

    k = Plan.from_dict({
        "target": t.to_dict(),
        "name": "my plan"
    })

    assert k.name == "my plan"
    assert isinstance(k.target, Target)
    assert k.target.earliest == 10
    assert k.target.latest == 20

    assert k.to_dict() == {
        "name": "my plan",
        "target": {
            "earliest": 10,
            "latest": 20,
            "obj_type": "Target",
        },
        "obj_type": "Plan",
    }


def test_embedded_many_with_dict():
    class SpeciesSchema(schema.BasicSchema):
        scientific_name = fields.Str()
        weight = fields.Str()
        habitats = fields.Embedded(obj_cls="Habitat", many=True)
        related_species = fields.Embedded(obj_cls="Species", many=True)

    class Species(onto.models.base.Serializable):
        _schema_cls = SpeciesSchema

    class EndangeredSpeciesSchema(SpeciesSchema):
        pass

    class EndangeredSpecies(onto.models.base.Serializable):
        _schema_cls = EndangeredSpeciesSchema

    class HabitatSchema(schema.BasicSchema):
        habitat_name = fields.Str()

    class Habitat(onto.models.base.Serializable):
        _schema_cls = HabitatSchema

    forests = Habitat.new(habitat_name="Forests")
    grasslands = Habitat.new(habitat_name="Grasslands")

    jaguar = Species.new(
        scientific_name="Panthera onca",
        habitats=[forests, grasslands]
    )

    cold_high_mountains = Habitat.new(
        habitat_name="cold high mountains")

    snow_leopard = Species.new(
        scientific_name="Panthera uncia",
        habitats=[cold_high_mountains]
    )

    temperate = Habitat.new(habitat_name="Temperate")
    broadleaf = Habitat.new(habitat_name="Broadleaf")
    mixed_forests = Habitat.new(habitat_name="Mixed Forests")

    amur_leopard = EndangeredSpecies.new(
        scientific_name="Panthera pardus orientalis",
        weight="70 - 105 pounds",
        habitats=[temperate, broadleaf, mixed_forests],
        related_species={
            "jaguar": jaguar,
            "snow leopard": snow_leopard
        }
    )

    d = {'habitats': [{'habitatName': 'Temperate'},
                      {'habitatName': 'Broadleaf'},
                      {'habitatName': 'Mixed Forests'}],
         'relatedSpecies': {'jaguar': {'habitats': [{'habitatName': 'Forests'},
                                                    {
                                                        'habitatName': 'Grasslands'}],
                                       'relatedSpecies': [],
                                       'scientificName': 'Panthera onca',
                                       'weight': ''},
                            'snow leopard': {
                                'habitats': [{'habitatName': 'cold high '
                                                             'mountains'}],
                                'relatedSpecies': [],
                                'scientificName': 'Panthera uncia',
                                'weight': ''}},
         'scientificName': 'Panthera pardus orientalis',
         'weight': '70 - 105 pounds'}

    assert d == amur_leopard.to_dict()

    amur_leopard_deserialized = EndangeredSpecies.from_dict(d)

    assert amur_leopard_deserialized.to_dict() == d



from onto.models.mixin import Importable, Exportable, NewMixin
from onto.models.base import BaseRegisteredModel, Schemed
from onto.view_model import ViewModel, ViewModelMixin


class Mutation(BaseRegisteredModel,
               Schemed, Importable, NewMixin, Exportable):

    view_model_cls = None


class PatchMutation(Mutation):

    def __init__(self):
        super().__init__()

    # @classmethod
    # def mutate_create(cls, doc_id=None, data=None):
    #     obj = cls.domain_model_cls.new(
    #         doc_id=doc_id,
    #         with_dict=data)
    #     obj.save()

    @classmethod
    def mutate_patch(cls, doc_id=None, data=None):
        obj = cls.view_model_cls.get(doc_id=doc_id)
        assert isinstance(obj, ViewModelMixin)
        obj.update_vals(with_dict=data)
        obj.propagate_change()

    @classmethod
    def mutate_patch_one(cls, obj=None, data=None):
        if data:
            assert isinstance(obj, ViewModelMixin)
            obj.update_vals(with_dict=data)
            obj.propagate_change()



def test_get_variables():

    class A:

        def __init__(self, a: str):
            self.a = a

        def normal_f(self):
            return f"normal a evaluates to {self.a}"

    import inspect
    # NOTE: getclosurevars has bugs that are still not fixed yet
    # https://stackoverflow.com/questions/61964532/inspect-getclosurevars-classifies-attributes-as-global-or-unbound-how-to-distin
    v = inspect.getclosurevars(A.normal_f)



import contextlib

# Attribute
# - Serialization/Deserialization instructions
# - Validation
# - Precondition
# - Type
from collections import defaultdict

import typing
from functools import cached_property

from onto.common import _NA


class _ModelRegistry(type):
    """

    TODO: add exception handling when resolving classes
            that are destructed.

    Attributes:
    ===================
    _REGISTRY: dict
        key: name of the class
        value: class

    """

    _REGISTRY = {}
    _tree = defaultdict(set)
    _tree_r = defaultdict(set)

    def __new__(mcs, name, bases, attrs):
        new_cls = type.__new__(mcs, name, bases, attrs)
        if new_cls.__name__ in mcs._REGISTRY:
            raise ValueError(
                "Class with name {} is declared more than once. "
                    .format(new_cls.__name__)
            )
        mcs._REGISTRY[new_cls.__name__] = new_cls

        for base in bases:
            if issubclass(type(base), mcs):
                mcs._tree[base.__name__].add(new_cls.__name__)
                mcs._tree_r[new_cls.__name__].add(base.__name__)

        return new_cls

    @classmethod
    def get_registry(mcs):
        return dict(mcs._REGISTRY)

    @classmethod
    def _get_children_str(mcs, cls_name):
        return mcs._tree[cls_name].copy()

    @classmethod
    def _get_parents_str(mcs, cls_name):
        return mcs._tree_r[cls_name].copy()

    @classmethod
    def get_cls_from_name(mcs, obj_type_str):
        """ Returns cls from obj_type (classname string)
        obj_type must be a subclass of cls in current class/object
        """
        if obj_type_str in mcs._REGISTRY:
            return mcs._REGISTRY[obj_type_str]
        else:
            return None


class MonadPrependMixin:

    def __enter__(self):
        return self.as_root()

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class Monad:

    def __init__(self, decor):
        self.decor: DecoratorBase = decor

    @classmethod
    def _get_root(cls):
        return cls(decor=BranchHead())

    # def __mul__(cls, cls_b):
    #     import types
    # cls_r = types.new_class(
    #     name=cls.__name__ + cls_b.__name__,
    #     bases=(cls, cls_b),
    # )
    # return cls_r

    @property
    def properties(self):
        if isinstance(self, BranchHead):
            raise TypeError('use attrs.nothing to prevent duplicate attrs values')
        return self.decor

    @property
    def fget(self):
        return self.properties.make_fget(name=self.name)

    @property
    def fset(self):
        return self.properties.make_fset(name=self.name)

    @property
    def fdel(self):
        return self.properties.make_fdel(name=self.name)

    @staticmethod
    def get_decorator_cls(item):
        from onto.utils import camelize
        name = camelize(item, uppercase_first_letter=True)
        decorator_cls = _ModelRegistry.get_cls_from_name(obj_type_str=name)
        if decorator_cls is None:
            raise AttributeError(f'Unable to locate {name} in subclass of DecoratorBase (for decorator: {item})')
        return decorator_cls

    def __getattr__(self, item):
        decorator_cls = self.get_decorator_cls(item=item)
        decor = self.decor
        self_cls = self.__class__

        def output_wrapper(decorated):
            return self_cls(decor=decorated)

        try:
            f = decorator_cls.easy(decor=decor, output_wrapper=output_wrapper)
        except AttributeError:
            raise AttributeError(f'Unable to transform attribute with {item}')

        return f

    def descendant_of(self, cs):
        cs = set(self.get_decorator_cls(item=it) for it in cs)
        return self.properties.descendant_of(ancestors=cs)


class MonadContext(Monad, contextlib.ContextDecorator):
    stack = list()

    @classmethod
    def context(cls):
        return cls(decor=root_decor.get())

    def __enter__(self):
        self.stack.append(root_decor.set(self.decor))

    def __exit__(self, exc_type, exc_val, exc_tb):
        root_decor.reset(self.stack.pop())
        return False


#
#
# class CurSelfContext(contextlib.ContextDecorator):
#
#     stack = list()
#
#     def __init__(self, cur_self):
#         self.cur_self = cur_self
#
#     def __enter__(self):
#         self.stack.append( cur_self.set(self.cur_self) )
#
#     def __exit__(self, exc_type, exc_val, exc_tb):
#         cur_self.reset(self.stack.pop())
#         return False


class MarshmallowCapableBaseMixin:

    @staticmethod
    def _customize_field_cls(field_cls, methods):
        new_field_cls = type('tmp_field', (field_cls,), methods)
        return new_field_cls

        # if not method_name:
        #     method_name = unbound_method.__name__
        # bound_method = unbound_method.__get__(field_obj, field_obj.__class__)
        # setattr(field_obj, method_name, bound_method)

    @property
    def _marshmallow_field_constructor(self):
        def _constructor(_self):
            methods_overridden = dict(_self._marshmallow_field_override)
            field_cls = self._customize_field_cls(_self._marshmallow_field_cls, methods_overridden)
            field_obj = field_cls(
                **dict(_self._marshmallow_field_kwargs)
            )

            return field_obj

        return _constructor

    @property
    def _marshmallow_field_kwargs(self):
        yield from ()

    @property
    def _marshmallow_field_cls(self):
        from onto.mapper import fields
        return fields.Field

    @property
    def _marshmallow_field_override(self):
        yield from ()


class GraphqlCapableMixin:

    @property
    def _graphql_object_type(self):
        yield from ()


class DefaultDecoratorMixin(MarshmallowCapableBaseMixin, GraphqlCapableMixin):
    is_internal = False


import contextvars

root_decor = contextvars.ContextVar('root_decor', default=DefaultDecoratorMixin())
cur_self = contextvars.ContextVar('cur_self', default=None)


def whichever_order(li: list, operation):
    """
    Hack: try operation on all permutative orders of li until the one order does not trigger an error
    """
    n = len(li)
    from itertools import permutations
    errors = list()
    for t in permutations(li, n):
        try:
            new_list = list(t)
            return operation(new_list)
        except Exception as e:
            errors.append(e)
    else:
        raise Exception(f'All possible combinations failed {str(errors)}')


class DecoratorBase(metaclass=_ModelRegistry):
    """
    Can only create new state in self.
    Do not modify state of self.decorated

    """

    # @property
    # def origin(self):
    #     return cur_self.get()  # TODO: implement

    @classmethod
    def new(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    @classmethod
    def easy_property(decorator_cls, decor, output_wrapper):
        decorated = decorator_cls.new(decorated=decor)
        return output_wrapper(decorated)

    @classmethod
    def easy_callable(decorator_cls, decor, output_wrapper):
        """
        Gets either a property or the initializer of a class
        """

        def f(*args, **kwargs):
            decorated = decorator_cls.new(*args, decorated=decor, **kwargs)
            return output_wrapper(decorated)

        return f

    @classmethod
    def easy(cls, *args, **kwargs):
        # Implemented verbosely by purpose to show as an example how to extend this method
        # easy will be bound to Decorator if defining easy = Decorator.easy in subclass

        return cls.easy_callable(*args, **kwargs)

    def __init__(self, decorated):
        super().__init__()
        self._decorated = decorated

    @property
    def decorated(self):
        decorated = self._decorated
        if isinstance(decorated, contextvars.ContextVar):
            decorated = decorated.get()
        return decorated

    def __getattr__(self, item):
        return getattr(self.decorated, item)

        # yield from self.decorated._marshmallow_field_kwargs

    def descendant_of(self, ancestors: typing.Set['DecoratorBase']):

        if self.__class__ in ancestors:
            # Trying to copy the set
            new_ancestor = ancestors - set((self.__class__,))
            ancestors = new_ancestor

        if not hasattr(self.decorated, 'descendant_of'):
            # TODO: risky, skipped for AttributeError occurring in attribute getter
            return len(ancestors) == 0
        else:
            return self.decorated.descendant_of(ancestors)

    import_required = False
    export_required = False

    is_root = False

    @property
    def graphql_object_type(self):
        typ = list(self._graphql_object_type)
        assert len(typ) != 0

        def op(typ):
            base = typ[0]
            for elem in typ[1:]:
                base = elem(base)
            return base

        return whichever_order(typ, op)


# _vals_

# class BindClass_(DecoratorBase):
#
#     @property
#     def parent(self):
#         return self._bound_cls
#
#     @property
#     def name(self):
#         return self.__attr_name
#
#     def __init__(self, *args, _bound_cls, _attr_name, **kwargs):
#         self.__bound_cls = _bound_cls
#         self.__attr_name = _attr_name
#         super().__init__(*args, **kwargs)
#
#     @property
#     def _bound_cls(self):
#         return self.__bound_cls


class Annotate(DecoratorBase):

    def __init__(self, annotation, *args, decorated, **kwargs):
        if annotation is not None:
            import typing
            if not isinstance(annotation, typing.Type):
                # annotation is a type alias
                # 2 layers max
                if origin := typing.get_origin(annotation):
                    if origin is list:
                        if arguments := typing.get_args(annotation):
                            element_type = next(iter(arguments))
                            if element_type in (str, bool, float, int,):
                                decorated = List(value=lambda a: a.of_type(element_type), decorated=decorated)
                    else:
                        decorated = OfType(type_cls=origin, decorated=decorated)
            else:
                decorated = OfType(type_cls=annotation, decorated=decorated)

        super().__init__(*args, decorated=decorated, **kwargs)


class BranchHead(DecoratorBase):

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)

    def __init__(self):
        super().__init__(decorated=root_decor)


class Nothing(DecoratorBase):
    """
    Use attr.nothing

    Wrong:
    class A:
      b: int = attr
      c: bool = attr
    A.b and A.c points to the same AttributeBase object, and
        b.name == c.name == 'c'
    """

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)


class DefaultValue(DecoratorBase):

    def __init__(self, default_value, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._default_value = default_value

    @property
    def default_value(self):
        return self._default_value

    @property
    def _marshmallow_field_kwargs(self):
        yield from self.decorated._marshmallow_field_kwargs
        yield 'missing', self.default_value


# class DefaultParamsMixin:
#     import_enabled = True
#     export_enabled = True


class ImportRequired(DecoratorBase):
    import_required = True

    @property
    def _marshmallow_field_kwargs(self):
        yield from self.decorated._marshmallow_field_kwargs
        yield ('required', True)

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)


class ExportRequired(DecoratorBase):
    # TODO: implement

    export_required = True

    @property
    def _marshmallow_field_kwargs(self):
        yield from self.decorated._marshmallow_field_kwargs
        yield ('required', True)

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)


class Required(ImportRequired, ExportRequired):

    @property
    def _graphql_object_type(self):
        import graphql
        # TODO: fix ordering
        yield graphql.GraphQLNonNull
        yield from self.decorated._graphql_object_type
    # pass


class Optional(DecoratorBase):
    import_required = False
    export_required = False

    @property
    def _marshmallow_field_kwargs(self):
        yield from self.decorated._marshmallow_field_kwargs
        yield 'required', False

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)


class ImportEnabled(DecoratorBase):
    import_enabled = True

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)


class ExportEnabled(DecoratorBase):
    export_enabled = True

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)


class Internal(DecoratorBase):
    is_internal = True
    import_enabled = False
    export_enabled = False

    @property
    def _marshmallow_field_kwargs(self):
        yield from self.decorated._marshmallow_field_kwargs
        yield 'dump_only', True
        yield 'load_only', True

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)


class DefaultParamsMixin(ImportEnabled, ExportEnabled, DecoratorBase):
    pass


class OfType(DecoratorBase):

    @classmethod
    def new(cls, *args, **kwargs):
        """
        Dispatch to subclass when required
        """
        target_cls = cls
        if args == (list,):  # TODO: make better
            target_cls = List
        return target_cls(*args, **kwargs)

    def __init__(self, type_cls, *args, **kwargs):
        self._type_cls = type_cls
        super().__init__(*args, **kwargs)

    @property
    def type_cls(self):
        return self._type_cls

    @property
    def _graphql_object_type(self):
        import graphql
        PY_TYPE_MAP_GQL = {
            int: graphql.GraphQLInt,
            bool: graphql.GraphQLBoolean,
            float: graphql.GraphQLFloat,
            str: graphql.GraphQLString,
            list: graphql.GraphQLList
        }
        t = self.type_cls
        if t in PY_TYPE_MAP_GQL:
            yield PY_TYPE_MAP_GQL[t]
        else:
            raise TypeError(f'Failed to locate graphql type for {t.__class__}')
        yield from self.decorated._graphql_object_type

    @property
    def _graphql_field_kwargs(self):
        yield from self.decorated._graphql_field_kwargs
        yield 'type_', self.graphql_object_type

    @property
    def _marshmallow_field_cls(self):
        from onto.mapper import fields
        from marshmallow import fields as marshmallow_fields
        from datetime import date, time, datetime, timedelta
        PY_TYPE_MARSHMALLOW_FIELD = {
            str: fields.String,
            int: fields.Integer,
            float: fields.Float,
            bool: fields.Boolean,
            date: marshmallow_fields.Date,
            time: marshmallow_fields.Time,
            timedelta: marshmallow_fields.TimeDelta,
            datetime: marshmallow_fields.DateTime
        }
        t = self.type_cls

        if t in PY_TYPE_MARSHMALLOW_FIELD:
            return PY_TYPE_MARSHMALLOW_FIELD[t]
        else:
            return fields.Field
            # raise TypeError(f'Failed to locate marshmallow field def for {t}')


class Pony(DecoratorBase):
    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        self.is_pony = True
        super().__init__(*args, **kwargs)

    def _make_pony_attribute_cls(self):
        from pony.orm import Discriminator, Required, Optional
        if self.decorated.__class__.__name__ == "Discriminator":
            # TODO: make better
            return Discriminator
        if self.decorated.import_required:
            return Required
        else:
            return Optional

    # @lru_cache(maxsize=None)
    def _make_pony_attribute(self):
        column_name = self.decorated.data_key
        py_type = self.decorated.type_cls
        is_required = self.decorated.import_required
        _pony_attribute_cls = self._make_pony_attribute_cls()
        _pony_attribute = _pony_attribute_cls(
            py_type, is_required=is_required, column=column_name)
        return _pony_attribute

    @cached_property
    def _pony_attribute(self):
        return self._make_pony_attribute()

    @property
    def is_collection(self):
        return self._pony_attribute.is_collection


class String(OfType):

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, type_cls=str, **kwargs)


class Str(String):
    pass


class Integer(OfType):

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, type_cls=int, **kwargs)


class Int(Integer):
    pass


class IntegerTimestamp(Integer):
    _long_type = None

    @classmethod
    def _get_long_type(cls):
        if not cls._long_type:
            from graphql import GraphQLScalarType
            cls._long_type = GraphQLScalarType('Long')
        return cls._long_type

    @property
    def _graphql_object_type(self):
        yield self._get_long_type()
        yield from self.decorated._graphql_object_type


class Float(OfType):

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, type_cls=float, **kwargs)


class Bool(OfType):

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, type_cls=bool, **kwargs)


#
# import contextvars
# current_self = contextvars.ContextVar('current_self', default=list())  # TODO: note mutable default
# from contextlib import contextmanager
# @contextmanager
# def use_self(*, self):
#     current_self.get().append(self)
#     try:
#         yield None
#     finally:
#         current_self.get().pop()


_ATTRIBUTE_STORE_NAME = '_attrs'


class Getter(DecoratorBase):

    def _get_default_fget(self, *, name):
        def fget(_self_obj):
            inner = getattr(_self_obj, _ATTRIBUTE_STORE_NAME)
            return getattr(inner, name)

        return fget

    def __init__(self, fget=_NA, *args, **kwargs):
        self._fget = fget
        super().__init__(*args, **kwargs)

    def make_fget(self, name):
        if self._fget is _NA:
            return self._get_default_fget(name=name)
        else:
            return self._fget


class Setter(DecoratorBase):

    def _get_default_fset(self, *, name):
        def fset(_self_obj, value):
            inner = getattr(_self_obj, _ATTRIBUTE_STORE_NAME)
            return setattr(inner, name, value)

        return fset

    def __init__(self, fset=_NA, *args, **kwargs):
        self._fset = fset
        super().__init__(*args, **kwargs)

    def make_fset(self, name):
        if self._fset is _NA:
            return self._get_default_fset(name=name)
        else:
            return self._fset


class Deleter(DecoratorBase):

    def _get_default_fdel(self, *, name):
        def fdel(_self_obj):
            inner = getattr(_self_obj, _ATTRIBUTE_STORE_NAME)
            return delattr(inner, name)

        return fdel

    def __init__(self, fdel=_NA, *args, **kwargs):
        self._fdel = fdel
        super().__init__(*args, **kwargs)

    def make_fdel(self, name):
        if self._fdel is _NA:
            return self._get_default_fdel(name=name)
        else:
            return self._fdel


class InitOptions(DecoratorBase):

    # @property
    # def initializer(self):
    #     return self._initializer

    def __init__(self, initialize, initializer, *args, **kwargs):
        self.initialize = initialize
        self._initializer = initializer
        super().__init__(*args, **kwargs)

    def make_init(self, name=_NA):
        return self._initializer


class Init(InitOptions):

    def __init__(self, initializer=_NA, *args, **kwargs):
        super().__init__(*args, initialize=True, initializer=initializer, **kwargs)


class EasyInit(Init):

    def make_init(self, name=_NA):
        def _init(_self_obj):
            inner = getattr(_self_obj, _ATTRIBUTE_STORE_NAME)
            return setattr(inner, name, self._easy_initializer(_self_obj))

        return _init

    def __init__(self, easy_initializer, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._easy_initializer = easy_initializer


class Dict(Init):

    def __init__(self, *args, **kwargs):
        def _dict_initializer(_self):
            attr_name = self.name  # TODO: fix
            setattr(_self, attr_name, dict())

        super().__init__(
            *args,
            initializer=_dict_initializer,
            **kwargs)


class List(OfType):

    def __init__(self, value, *args, **kwargs):
        self._list_value = value
        super().__init__(*args, type_cls=list, **kwargs)

    @property
    def list_value(self):
        return self._list_value

    @property
    def _marshmallow_field_kwargs(self):
        yield from self.decorated._marshmallow_field_kwargs
        # TODO: fix peer
        ins = self.list_value.properties._marshmallow_field_constructor(self.list_value.properties)
        yield ('cls_or_instance', ins)

    @property
    def _marshmallow_field_cls(self):
        from onto.mapper import fields
        return fields.List

    @property
    def _graphql_object_type(self):
        import graphql
        yield graphql.GraphQLList(
            self.list_value.properties.graphql_object_type
        )
        yield from self.decorated._graphql_object_type


class Enum(OfType):
    from enum import Enum as _Enum

    def __init__(self, enum_cls: typing.Type[_Enum], *args, **kwargs):
        self._enum_cls = enum_cls
        super().__init__(*args, type_cls=str, **kwargs)

    @property
    def _graphql_object_type(self):
        import graphql
        yield graphql.GraphQLEnumType(
            self._enum_cls.__name__,
            values=self._enum_cls
        )
        yield from self.decorated._graphql_object_type


class AttributeName(DecoratorBase):

    @property
    def _marshmallow_field_kwargs(self):
        yield from self.decorated._marshmallow_field_kwargs
        yield ('attribute', self.name)

    def __init__(self, name, *args, **kwargs):
        self.name = name
        super().__init__(*args, **kwargs)


class ParentKlass(DecoratorBase):

    def __init__(self, parent, *args, **kwargs):
        self.parent = parent
        super().__init__(*args, **kwargs)


class DocId(DecoratorBase):

    # @classmethod
    # def new(cls, *args, **kwargs):
    #     """
    #     Dispatch to subclass when required
    #     """
    #
    #     def fget(self):
    #         return self.doc_ref.id
    #
    #     inner = Getter(fget, *args, **kwargs)
    #     return cls(decorated=inner)

    @property
    def _graphql_object_type(self):
        import graphql
        yield graphql.GraphQLNonNull(graphql.GraphQLID)
        yield from self.decorated._graphql_object_type

    @property
    def _graphql_field_kwargs(self):
        yield from self.decorated._graphql_field_kwargs
        yield 'type_', self.graphql_object_type

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)

    @property
    def _marshmallow_field_cls(self):
        from onto.mapper import fields
        return fields.DocIdField


class NodeId(DecoratorBase):
    """
    For GraphQL ID
    """

    @property
    def _graphql_object_type(self):
        import graphql
        yield graphql.GraphQLNonNull(graphql.GraphQLID)
        yield from self.decorated._graphql_object_type

    @property
    def _graphql_field_kwargs(self):
        yield from self.decorated._graphql_field_kwargs
        yield 'type_', self.graphql_object_type

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_property(*args, **kwargs)


class DataKey(DecoratorBase):

    @property
    def _marshmallow_field_kwargs(self):
        yield from self.decorated._marshmallow_field_kwargs
        yield ('data_key', self.data_key)

    @property
    def data_key(self):
        data_key = self._data_key
        if data_key is not None:
            return self._data_key
        else:
            raise AttributeError

    def __init__(self, data_key=_NA, *args, **kwargs):
        self._data_key = data_key
        super().__init__(*args, **kwargs)


class DataKeyFromName(DataKey):

    @classmethod
    def easy(cls, *args, **kwargs):
        return cls.easy_callable(*args, **kwargs)

    @property
    def data_key(self):
        return self._transformation(self.name)

    def __init__(self, *args, transformation=_NA, **kwargs):
        if transformation is _NA:
            from onto.utils import camel
            transformation = camel
        self._transformation = transformation
        super().__init__(*args, **kwargs)


class NoneAsMissing(DecoratorBase):

    @property
    def _marshmallow_field_override(self):
        yield from self.decorated._marshmallow_field_override

        def _deserialize(_self, value, attr, data, **kwargs):
            if value is None:
                from marshmallow.fields import missing_
                return missing_
            else:
                return super(_self.__class__, _self)._deserialize(value, attr, data, **kwargs)

        yield ('_deserialize', _deserialize)

    @property
    def _marshmallow_field_kwargs(self):
        yield from self.decorated._marshmallow_field_kwargs
        yield 'allow_none', True


class Embed(OfType):

    @property
    def _marshmallow_field_kwargs(self):
        yield from self.decorated._marshmallow_field_kwargs
        yield 'obj_type', self.type_cls
        yield 'allow_none', True  # TODO: make better

    @property
    def _marshmallow_field_cls(self):
        from onto.mapper import fields
        return fields.Embedded

    @property
    def _graphql_object_type(self):
        type_cls = self.type_cls
        if isinstance(type_cls, str):
            from onto.models.meta import ModelRegistry
            type_cls = ModelRegistry.get_cls_from_name(obj_type_str=type_cls)
        # from onto.models.utils import _graphql_object_type_from_attributed_class
        yield type_cls.get_graphql_object_type(is_input=self.is_input)
        yield from self.decorated._graphql_object_type

    # @property
    # def _graphql_(self):
    #     e_cls = attr.type_cls
    #     e_graphql = _graphql_object_type_from_attributed_class(e_cls)
    #     field = field_base(
    #         type_=f(e_graphql),
    #         description=attr.doc
    #     )
    #     return attr.data_key, field


class Set(DecoratorBase):

    def __init__(self, value, *args, **kwargs):
        self.value = value
        super().__init__(*args, **kwargs)


class ReferenceSet(Set):

    def __init__(self, value, *args, **kwargs):
        self.value = value
        super().__init__(*args, **kwargs)

    @property
    def _pony_field(self):
        raise NotImplementedError


class ReferenceList(List):
    pass


class WithReference(DecoratorBase):
    pass


class Relation(DecoratorBase):

    @property
    def _marshmallow_field_kwargs(self):
        yield from self.decorated._marshmallow_field_kwargs
        yield ('dm_cls', self.dm_cls)

    def __init__(self, dm_cls, *args, **kwargs):
        self.dm_cls = dm_cls
        super().__init__(*args, **kwargs)

    @property
    def _marshmallow_field_cls(self):
        from onto.mapper import fields
        return fields.Relationship


class GraphqlCapable(DecoratorBase):

    def __init__(self, *args, is_input, **kwargs):
        self.is_input = is_input
        super().__init__(*args, **kwargs)

    @property
    def _graphql_field_kwargs(self):
        yield from ()


class Doc(DecoratorBase):
    """
    Documented with str
    """

    def __init__(self, doc, *args, **kwargs):
        self._doc = doc
        super().__init__(*args, **kwargs)

    @property
    def doc(self):
        return self._doc

    @property
    def _graphql_field_kwargs(self):
        yield from self.decorated._graphql_field_kwargs
        yield 'description', self.doc


class Executable(Getter):
    """
    可以被执行的 attribute 用于 graphql
    """

    def __init__(self, f=_NA, *args, **kwargs):
        self._f = f
        super().__init__(lambda *_: f, *args, **kwargs)

    def _make_graphql_args(self):
        from onto.sink.graphql import GraphQLSink
        parameters = GraphQLSink._parameters_for(self._f)
        for name, annotated_type in parameters:
            yield name, GraphQLSink._param_to_graphql_arg(annotated_type=annotated_type)
        yield from ()

    @property
    def _graphql_field_kwargs(self):
        yield from self.decorated._graphql_field_kwargs
        yield 'args', dict(self._make_graphql_args())


class AsRoot(DecoratorBase):
    is_root = True



from collections import OrderedDict

import pytest

import onto.models.utils
from onto import schema
from onto.attrs import attribute
from onto.models.base import Serializable


class G(Serializable):
    i = attribute.AttributeBase()


def test__collect_attrs():
    assert isinstance(G.i, attribute.AttributeBase)

    assert list(onto.models.utils._collect_attrs(G)) == [("i", G.i)]


def test__schema_cls_from_attributed_class():
    res = onto.models.utils._schema_cls_from_attributed_class(G)

    assert isinstance(res, schema.Schema.__class__)
    schema_obj: schema.Schema = res()
    assert isinstance(schema_obj.fields, OrderedDict)
    print(schema_obj.fields)

    # desc = "OrderedDict([('obj_type', <fields.Function(" \
    #        "default=<marshmallow.missing>, attribute='obj_type', " \
    #        "validate=None, required=False, load_only=False, dump_only=False, " \
    #        "missing=<marshmallow.missing>, allow_none=False, " \
    #        "error_messages={'required': 'Missing data for required field.', " \
    #        "'null': 'Field may not be null.', 'validator_failed': 'Invalid " \
    #        "value.'})>), ('doc_id', <fields.String(" \
    #        "default=<marshmallow.missing>, attribute='doc_id', " \
    #        "validate=None, required=False, load_only=False, dump_only=True, " \
    #        "missing=<class 'str'>, allow_none=False, error_messages={" \
    #        "'required': 'Missing data for required field.', 'null': 'Field " \
    #        "may not be null.', 'validator_failed': 'Invalid value.', " \
    #        "'invalid': 'Not a valid string.', 'invalid_utf8': 'Not a valid " \
    #        "utf-8 string.'})>), ('doc_ref', <fields.String(" \
    #        "default=<marshmallow.missing>, attribute='doc_ref_str', " \
    #        "validate=None, required=False, load_only=False, dump_only=True, " \
    #        "missing=<class 'str'>, allow_none=False, error_messages={" \
    #        "'required': 'Missing data for required field.', 'null': 'Field " \
    #        "may not be null.', 'validator_failed': 'Invalid value.', " \
    #        "'invalid': 'Not a valid string.', 'invalid_utf8': 'Not a valid " \
    #        "utf-8 string.'})>), ('_remainder', <fields.Remainder(" \
    #        "default=<marshmallow.missing>, attribute='_remainder', " \
    #        "validate=None, required=False, load_only=False, dump_only=False, " \
    #        "missing=<marshmallow.missing>, allow_none=False, " \
    #        "error_messages={'required': 'Missing data for required field.', " \
    #        "'null': 'Field may not be null.', 'validator_failed': 'Invalid " \
    #        "value.', 'invalid': 'Not a valid mapping type.'})>), ('i', " \
    #        "<fields.Field(default=<marshmallow.missing>, attribute='i', " \
    #        "validate=None, required=False, load_only=False, dump_only=False, " \
    #        "missing=None, allow_none=True, error_messages={'required': " \
    #        "'Missing data for required field.', 'null': 'Field may not be " \
    #        "null.', 'validator_failed': 'Invalid value.'})>)]) "
    # assert str(schema_obj.fields) == desc

